#include "ocsinv.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <zlib.h>


int debug = 0;
int local = 0; /* 0=non local mode, 1=dump uncompressed XML to stdout, 2=dump compressed XML to file */
int auto_update = 0;
char* ocs_server = NULL;
char  *deviceid_dir = NULL;
char *myname = NULL;

static int usage(char* pg)
{
	fprintf(stdout, "%s: OCS client for Solaris (%s/%s)\n"
			"\n"
			"Usage: %s [-l] [-d] -s <server>:[<port>] [-a]\n\n"
			"Where:\n"
			"  -l           local mode (display XML document)\n"
			"  -L           local mode (create .ocs file)\n"
			"  -d           debug mode\n"
			"  -s <server>  name of the OCS server\n"
			"  -V           print agent version and exit\n"
			"  -a           activate auto-update (default: OFF)\n"
			"  -f <file>    file or directory where to put devide-id file\n"
			"  -n <name>    name to appear on server\n"
			"\n\n", pg, AGENT_NAME, AGENT_VERSION, pg);

	exit(0);
}


/*
 * dump the XML document to a compressed file (<deviceid>.ocs)
 *
 */

#define CHUNK 16384

int dumpOCSFile_write_cb (void* ctxt, const char* data, int len)
{
	z_stream* zstrm = (z_stream*) ctxt;
	unsigned char out[CHUNK];
	int inpos;
	int err;
	int n;

	inpos=0;

	do
	{
		n  = ((len-inpos) > CHUNK ) ? CHUNK : (len-inpos);
		zstrm->avail_in = n;
		zstrm->next_in = (char*) data + inpos;

		do {	
			zstrm->next_out = out;
			zstrm->avail_out = CHUNK;
			err = deflate ( zstrm, Z_NO_FLUSH );

			if (err!=Z_OK) { LOG("%s: %i", "deflate() problem.", err); return -1;}

			DEBUG("ecriture de %i bytes comp.", CHUNK - zstrm->avail_out);
			if ((CHUNK - zstrm->avail_out))
				fwrite ( out , 1, CHUNK - zstrm->avail_out,  (FILE*) zstrm->opaque );

		} while (( zstrm->avail_out == 0) && (err == Z_OK));
		
		inpos +=n; 
	} while (inpos < len);

	return len;
}

int dumpOCSFile_close_cb (void* ctxt)
{
	z_stream* zstrm = (z_stream*) ctxt;
	unsigned char out[CHUNK];
	int err;

	do 
	{
		zstrm->next_out = out;
		zstrm->avail_out = CHUNK;
		zstrm->avail_in = 0;
		zstrm->next_in  = Z_NULL;

		err = deflate ( zstrm, Z_FINISH);
		if ((err == Z_OK) || (err = Z_STREAM_END))
			fwrite ( out , 1, CHUNK - zstrm->avail_out, (FILE*) zstrm->opaque );
	} while ( (zstrm->avail_out == 0) && ( err == Z_OK));

	fclose ((FILE*) zstrm->opaque);
}


static void dumpOCSFile (const xmlDocPtr invdoc)
{
	xmlOutputBufferPtr	outbuf;
	z_stream zstrm;
	char filename[PATH_MAX];
	int err;
	FILE* zfile;

	zstrm.zalloc = Z_NULL;
	zstrm.zfree = Z_NULL;
	zstrm.opaque = Z_NULL;
	zstrm.avail_in = 0;
	zstrm.next_in = Z_NULL;

	err = deflateInit(&zstrm, Z_DEFAULT_COMPRESSION);

	if (err!= Z_OK)
	{
		LOG("%s", "compression init. failed.");
		return ;
	}

	snprintf (filename, PATH_MAX, "%s.ocs", get_device_id());
	zfile = fopen ( filename, "wb");
	if (!zfile)
	{
		LOG("%s: %s", filename, "Can create file");
		return;
	}

	zstrm.opaque = zfile;
	
	outbuf = xmlOutputBufferCreateIO ( dumpOCSFile_write_cb, dumpOCSFile_close_cb, &zstrm, NULL);
	if (!outbuf)
	{
		LOG("%s: %s", filename, "Could not create output buffer (short memory?)");
		fclose (zfile);
		deflateEnd (&zstrm);
		return;
	}

	err = xmlSaveFileTo ( outbuf, invdoc, NULL);
	if (err == -1)
		LOG("%s: %s", filename, "Could not save document in file.");

	fclose (zfile);	
	deflateEnd (&zstrm);

	LOG("Created file %s for OCS import.", filename);

}

int main(int argc, char** argv)
{
	xmlDocPtr invdoc;
	xmlNodePtr root, node, subnode;
	OCS_t* ocs;
	int opt;

	/* init */

	xmlKeepBlanksDefault (0); 
	ocs_server = NULL;

	/* parse command line options */

	while ( ((opt=getopt(argc, argv, "lLds:Vaf:n:")) != EOF))
	{
		switch (opt)
		{
			case 'l':
				local=1;
				break;

			case 'L':
				local=2;
				break;

			case 'd':
				debug =1;
				break;

			case 's':
				ocs_server = optarg;
				break;

			case 'V':
				fprintf(stdout, "%i\n", AGENT_VERSION_REV);
				exit (EXIT_SUCCESS);

			case 'a':
				auto_update =1;
				break;

			case 'f':
				deviceid_dir = optarg;
				break;

			case 'n':
				myname = optarg;
				break;

			default:
				usage(argv[0]);
		}
	}

	if (!(ocs_server) && (!local)) usage(argv[0]);

	if (!local)
	{
		int new_rev;
		int res;

		ocs = OCS_new(ocs_server);

		assert (ocs != NULL);

		DEBUG ("OCS server: '%s'", ocs_server);	

		if (auto_update)
		{

			/* check if a new agent is available for download */

			LOG ("Check agent update... (current revision: %i) ...", AGENT_VERSION_REV);
			new_rev = OCS_check_agent_update(ocs);	
			DEBUG ("... OCS_check_agent_update() returned %i", new_rev);

			if (new_rev >0)
			{
				LOG ("New agent available for download (revision %i) ...", new_rev);

				if (OCS_update_agent (ocs, new_rev))
				{
					LOG("%s", "Update failed...");
				}
				else
				{
					char** _args;
					int n;

					LOG("%s", "New agent sucessfully downloaded & installed. Restarting the agent...");

					_args= (char**) malloc ( sizeof(char*) * (argc+1));
					for (n=0; n < argc;  n++) _args[n] = strdup (argv[n]);
					_args[argc+1]=NULL;

					if  ( execv (_args[0], _args)!=0)
					{
						perror(_args[0]);
						exit (EXIT_FAILURE);
					}
				}
			}
			else
					LOG("%s", "No update available ...");
		}


		/* check if we need to send the inventory */

		DEBUG("%s", "Check if we need to send the inventory ...");
		res = OCS_check_inventory (ocs);
		DEBUG("... OCS_check_inventory() returned %i", res);

		if (res==0)
		{
			LOG("OCS server '%s' rejects our inventory update ...", ocs_server);
			OCS_free (ocs);
			exit (EXIT_SUCCESS);
		}
	}

	/* setup the inventory document */

	invdoc = OCS_new_doc ("INVENTORY");
	root= xmlDocGetRootElement(invdoc);

	/* content subnode */

	node = xmlNewChild (root, NULL, "CONTENT", NULL);
	info_content (node);


	/* hardware subnode */

	subnode = xmlNewChild(node, NULL, "HARDWARE", NULL);
	if (subnode)
	{
		info_os     (subnode);
		info_memory (subnode);
	}

	/* BIOS sunode */

	info_bios(node);

	/* network subnode */

	info_network (node);

	/* drive subnode */

	info_drives (node);

	/* storage subnode */
	info_storage (node);

	/* software subnode */
	info_software (node);

	switch (local)
	{
		case 1:
			xmlDocFormatDump (stdout, invdoc, 1);
			break;

		case 2:
			dumpOCSFile (invdoc);
			break;

		default:
			{	
				int res;

				LOG("Sending inventory document to %s ...", ocs_server);    
				res = OCS_send_inventory (ocs, invdoc);
				DEBUG ("... OCS_send_inventory() returned %i", res);

				if (res!= 0)
					LOG("... failed to send the inventory to %s", ocs_server);

				xmlFreeDoc (invdoc);    
				OCS_free (ocs);
			}
	}

	return 0;
}

