#include <unistd.h>
#include <zlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>


#include "ocsinv.h"


static int debug_print_cb (CURL * curl, curl_infotype type, char * msg, size_t len, void *x );


/*
 * Simple XPath-like node resolver. 
 *
 */

static xmlNodePtr xmlSimpleQuery(const xmlDocPtr xmldoc, const char* _expr)
{

	char buf[256], *expr, *next ;
	xmlNodePtr node, child;
	int found;

	snprintf (buf, sizeof(buf), "%s", _expr);
	expr = buf;
	node = NULL;

	do 
	{
		next = strpbrk (expr, "/");

		if (next) {		
			*next ='\0';
			next++;
		}

		if (!node)
		{
			node = xmlDocGetRootElement (xmldoc);
			if (!node || (xmlStrcmp (node->name, expr) != 0))
				return NULL;

		}
		else
		{
			child = node->children;
			found = 0;
			while (child)
			{
				if ( xmlStrcmp (child->name, expr)==0)
				{
					node=child;
					found = 1; break;
				}
				child=child->next;
			}
			if (!found) return NULL; /* not found */
		}				
		expr = next;
	} while (expr);		
	return node;
}





/*
 * returns a OCS Request XML document pointer.
 *
 */

	xmlDocPtr
OCS_new_doc (const char* query)
{
	xmlDocPtr doc;
	xmlNodePtr root;

	doc = xmlNewDoc (XML_VERSION);
	if (!doc)
		return NULL;

	root= xmlNewNode (NULL, "REQUEST");
	xmlDocSetRootElement (doc, root);
	xmlNewTextChild (root, NULL, "QUERY", query);	

	return doc;

}



/*
 * returns 1 if we have to send the inventory,
 *         0 if not,
 *        -1 if an error occured
 */

	int       
OCS_check_inventory(OCS_t* me)
{
	xmlDocPtr prolog_doc, resp;
	xmlNodePtr root, node;
	xmlChar* content;
	int update_required;

	
	prolog_doc = OCS_new_doc ("PROLOG");
	assert (prolog_doc != NULL);

	root = xmlDocGetRootElement (prolog_doc);
	xmlNewTextChild (root, NULL, "DEVICEID", get_device_id());    

	resp = OCS_request (me, prolog_doc);

	xmlFreeDoc (prolog_doc);

	if (!resp)
	{
		LOG("** ERROR: OCS server '%s' returned unparsable response !", me->server);
		xmlFree (prolog_doc);
		return -1;
	}

	if (debug)
	{
		DEBUG("Dumping PROLOG response from '%s'", me->server);
		xmlDocFormatDump (stdout, resp, 1);
	}

	node = xmlSimpleQuery (resp, "REPLY/RESPONSE");
	if (!node || (!(content = xmlNodeGetContent (node))))
	{
		LOG("** ERROR: OCS server '%s' returned invalid response ! ", me->server);
		xmlFree (prolog_doc); xmlFree (resp);			
		return -1;
	}

	update_required= (xmlStrcmp(content, "SEND")==0);
	xmlFree (content);
	xmlFreeDoc (resp);

	DEBUG("Update inventory required ? %s", update_required ? "YES" : "NO");

	return update_required;
}


/*
 * Check if a new agent is available for download.
 *
 * Returns : 
 *   n>0  the new agent revision number.
 *     0  no update required.
 *    -1  if an error occured.
 */


	int       
OCS_check_agent_update (OCS_t* me)
{
	xmlDocPtr update_doc, resp;
	xmlNodePtr root, node;
	xmlChar* content;	
	char buf[256];
	int new_rev;

	snprintf(buf, 256, "%i", AGENT_VERSION_REV);	
	DEBUG("Check if a new agent is available... (current agent revision=%i)", AGENT_VERSION_REV);

	update_doc = OCS_new_doc ("UPDATE");
	assert (update_doc != NULL);

	root = xmlDocGetRootElement (update_doc);

	xmlNewTextChild (root, NULL, "DEVICEID", get_device_id());    
	xmlNewTextChild (root, NULL, "PLATFORM", AGENT_PLATFORM );
	xmlNewTextChild (root, NULL, "AGENT",  buf); 

	resp = OCS_request (me, update_doc);
	if (!resp)
	{
		LOG("** ERROR: OCS server '%s' returned unparsable response !", me->server);
		xmlFreeDoc (update_doc);
		return -1;			
	}

	if (debug)
	{
		DEBUG("Dumping UPDATE response from '%s'", me->server);
		xmlDocFormatDump (stdout, resp, 1);
	}

	node = xmlSimpleQuery (resp, "REPLY/RESPONSE");
	if (!node || (!(content = xmlNodeGetContent (node))))
	{
		LOG("** ERROR: OCS server '%s' returned invalid response ! ", me->server);
		xmlFreeDoc (update_doc); xmlFreeDoc (resp);
		return -1;
	}

	new_rev = 0;

	if (xmlStrcmp (content, "UPDATE")==0)
	{
		xmlFree (content);
		node = xmlSimpleQuery (resp, "REPLY/AGENT");
		if (node && 
				(content = xmlNodeGetContent(node)) &&
				(new_rev = atoi (content)))
		{
			xmlFree (content); /* new_rev contains the new revision number */
		}
	}

	xmlFree (content);							
	xmlFreeDoc (resp);
	xmlFreeDoc (update_doc);						
	return new_rev;
}

/*
 * send the inventory to the OCS server.
 *
 * Returns O if ok, -1 if not 
 *
 */

	int	
OCS_send_inventory (OCS_t* me, const xmlDocPtr inventory_doc)
{
	xmlDocPtr resp;


	resp = OCS_request (me, inventory_doc);

	if (resp)
	{
		if (debug)
		{
			DEBUG("%s", "Dumping response ...");
			xmlDocFormatDump (stdout, resp,  1);
		}
		xmlFreeDoc (resp);

		return 0;
	}

	return -1;

}

/*
 * download, uncompress and install a new agent.
 *
 * Returns 0 if ok, -1 if not.
 *
 */

	int 
OCS_update_agent (OCS_t* me, int revision)
{
	char URL[256];

	FILE*  tmpfile;
	char   tmpfile_name[256];
	double tmpfile_size;

	FILE*   newfile;
	char    newfile_name[256];

#define Z_CHUNK 4096	
	char    buf_in [Z_CHUNK];
	char    buf_out[Z_CHUNK];	

	char    cmd[256];
	int     new_rev;

	char    symlink_name[256];

	int      ret;
	z_stream zstrm;

	char* env_tmp;
	int curl_res;


	/* create temp. file */

	snprintf(tmpfile_name, sizeof (tmpfile_name), 
			"%s/%s-temp.XXXXXX",
			(env_tmp=getenv("TEMP")) ? env_tmp : "/tmp",
			AGENT_NAME);

#ifdef HAVE_MKSTEMP
	if (!mkstemp(tmpfile_name))
#else
	if (!mktemp(tmpfile_name))
#endif
	{
		perror(tmpfile_name);
		return -1;
	}

	DEBUG ("using temporary file %s", tmpfile_name);


	tmpfile = fopen (tmpfile_name, "wb+");
	if (!tmpfile)
	{
		perror(tmpfile_name);
		return -1;
	}

	snprintf(URL, sizeof (URL), "http://%s//ocsinventory/update/%s/%s/%i", 
			me->server,
			AGENT_PLATFORM,
			AGENT_NAME,
			revision);

	DEBUG("downloading %s", URL);

	/* setup CURL options */
#if LIBCURL_VERSION_NUM > 0x070c01  
	curl_easy_reset (me->curl);	/* available since 7.12.1 */
#else
	curl_easy_cleanup (me->curl);
	me->curl = curl_easy_init ();
	assert (me->curl != NULL);
#endif
	curl_easy_setopt (me->curl, CURLOPT_URL, URL);    
	if (debug) curl_easy_setopt (me->curl, CURLOPT_VERBOSE, 1);
	curl_easy_setopt (me->curl, CURLOPT_DEBUGFUNCTION, debug_print_cb );
	curl_easy_setopt (me->curl, CURLOPT_USERAGENT, AGENT_NAME "/" AGENT_VERSION);  
	curl_easy_setopt (me->curl, CURLOPT_FOLLOWLOCATION, 1);
	curl_easy_setopt (me->curl, CURLOPT_HTTPGET, 1L);
	curl_easy_setopt (me->curl, CURLOPT_WRITEFUNCTION, NULL);	
	curl_easy_setopt (me->curl, CURLOPT_WRITEDATA, tmpfile);		
	/* proposed by domfrance: set connection timeout. */
	curl_easy_setopt (me->curl, CURLOPT_CONNECTTIMEOUT, OCS_CONNECT_TIMEOUT); 
	curl_easy_setopt (me->curl, CURLOPT_TIMEOUT, OCS_TIMEOUT); 

	curl_res = curl_easy_perform (me->curl);

	if (curl_res != 0)
	{
		fclose (tmpfile);	
		unlink (tmpfile_name);

#if LIBCURL_VERSION_NUM > 0x070c00
                LOG("** ERROR: Can't download %s: %s", URL, curl_easy_strerror(curl_res));
#else
                LOG("** ERROR: Can't download %s: error %i", URL, curl_res);
#endif
		return -1;
	}

	tmpfile_size = (double) 0.0;
	curl_easy_getinfo (me->curl, CURLINFO_SIZE_DOWNLOAD, &tmpfile_size);

	if (! tmpfile_size)
	{
		fclose (tmpfile);		
		unlink (tmpfile_name);	
		LOG("** ERROR: Empty download (%s)", URL);
		return -1;
	}

	LOG("New agent downloaded from %s, compressed file size is %.0f bytes.", URL, tmpfile_size);

	snprintf(newfile_name, sizeof (newfile_name), "%s/%s.%i", 
			BINDIR,
			AGENT_NAME, 
			revision);

	DEBUG("Will uncompress new agent in %s", newfile_name);

	if (access (newfile_name, F_OK)==0)
	{
		fclose (tmpfile);		
		LOG("** ERROR: can not replace existing file %s.", newfile_name);
		unlink (tmpfile_name);
		return -1;
	}

	fseek (tmpfile, 0L, SEEK_SET);	

	newfile = fopen (newfile_name, "wb");
	if (!newfile)
	{
		fclose (tmpfile);		
		perror(newfile_name);
		unlink (tmpfile_name);		
		return -1;
	}

	zstrm.zalloc = Z_NULL;
	zstrm.zfree = Z_NULL;
	zstrm.opaque = Z_NULL;
	zstrm.avail_in = 0;
	zstrm.next_in = Z_NULL;
	ret = inflateInit(&zstrm);

	do 
	{
		zstrm.avail_in = fread ( buf_in, 1, Z_CHUNK, tmpfile);
		if (zstrm.avail_in <=0) break;
		zstrm.next_in = buf_in;

		do {
			zstrm.next_out = buf_out;		
			zstrm.avail_out = Z_CHUNK;

			ret = inflate(&zstrm, Z_NO_FLUSH);
			if (ret != Z_OK) break;

			fwrite (buf_out, 1, Z_CHUNK - zstrm.avail_out, newfile);				
		} while (zstrm.avail_out == 0);	
	} while ( ! feof (tmpfile));

	fclose (tmpfile);
	fclose (newfile);	
	chmod (newfile_name, 0755);			
	unlink (tmpfile_name);

	/*
	 * try to run this agent, and test revision number.
	 * Do not update the symbolic link if revision number is wrong.
	 */

	snprintf(cmd, sizeof (cmd), "%s -V", newfile_name);	 
	tmpfile = popen (cmd, "r");
	if (!tmpfile)
	{
		perror(cmd);
		unlink (newfile_name);
		return -1;
	}

	if (fscanf(tmpfile, "%i", &new_rev) != 1)
	{
		LOG("** ERROR: invalid output of %s", cmd);
	//	unlink (newfile_name);
		pclose (tmpfile);
		return -1;
	}

	if (new_rev != revision)
	{
		LOG("*** ERROR: downloaded file %s has wrong revision number (%i, expected %i)", 
				URL, new_rev, revision);
		unlink (newfile_name);
		pclose (tmpfile);
		return -1;
	}

	/* that's all good... */

	snprintf(symlink_name, sizeof(symlink_name), "%s/%s", BINDIR, AGENT_NAME);
	unlink (symlink_name);
	if (symlink (newfile_name, symlink_name) != 0)
	{	
		perror(symlink_name);
		return 1;
	}	

	return 0;

}

static size_t doc_compress (char* in, size_t in_len, char** out)
{
	size_t outlen;
	size_t outsize;
	int finish;
	char buf[Z_CHUNK];
	z_stream z_strm;
	int ret;
	int n;

	outlen = 0;
	/* provide a default output buffer */
	outsize=Z_CHUNK;
	*out = (char*) malloc (outsize);
	assert (*out != NULL);

	memset(&z_strm, 0, sizeof (z_stream));
	z_strm.zalloc = Z_NULL;
	z_strm.zfree = Z_NULL;
	z_strm.opaque = Z_NULL;

	ret = deflateInit (&z_strm, Z_DEFAULT_COMPRESSION);
	assert (ret == Z_OK);

	z_strm.next_in = in;
	z_strm.avail_in =  in_len;

	do {

		do {

			z_strm.next_out = buf;
			z_strm.avail_out = Z_CHUNK;

			finish = (z_strm.avail_in) ? Z_NO_FLUSH : Z_FINISH;

			ret = deflate (&z_strm, finish);
			assert(ret != Z_STREAM_ERROR);

			n = Z_CHUNK - z_strm.avail_out; /* n encoded */

			/* realloc out buffer, if needed */
			while (outsize < (outlen + n))
			{
				outsize*=2;
				*out = (char*) realloc (*out, outsize);
				assert (*out != NULL);
			}
			memcpy (*out+outlen, buf, n);
			outlen +=n;

		} while ( z_strm.avail_out == 0);
	}while (finish != Z_FINISH);

	deflateEnd (&z_strm);  
	return outlen;
}

static size_t write_doc_cb( void *ptr, size_t size, size_t nmemb, void *stream)
{
	z_stream *z_strm ;
	xmlParserCtxtPtr parser = (xmlParserCtxtPtr) stream;
	char buf[Z_CHUNK];
	int ret;

	assert (parser->_private != NULL);

	z_strm = (z_stream*) parser->_private;

	z_strm->next_in = ptr;
	z_strm->avail_in = size * nmemb;

	do 
	{
		memset(buf, 0, Z_CHUNK);

		z_strm->next_out = buf;
		z_strm->avail_out = Z_CHUNK;

		ret = inflate (z_strm, Z_NO_FLUSH);     

		xmlParseChunk (parser, buf, Z_CHUNK - z_strm->avail_out, 0);
		if (ret != Z_OK) return size * nmemb; /* end or file or error */      

	} while (z_strm->avail_in != 0);

	return size * nmemb;
}

	static int 
debug_print_cb (CURL * curl, curl_infotype type, char * msg, size_t len, void *x )
{

	/* do not display row DATA: this is compressed XML document */

	switch (type)
	{
		case CURLINFO_TEXT:
#if 0
		case CURLINFO_HEADER_IN:
		case CURLINFO_HEADER_OUT:
#endif    
		case CURLINFO_END:
			fprintf(stderr, "** [CURL (%i)]:  %s", type, msg);
			break;

		default:
			break;
	}


	return 0;
}


OCS_t* OCS_new (const char* ocs_server)
{
	OCS_t* me;

	me = (OCS_t*) malloc (sizeof(OCS_t));
	assert (me != NULL);

	memset (me,0, sizeof (OCS_t));

	me->server = strdup (ocs_server);
	me->curl = curl_easy_init();
	assert (me->curl != NULL);

	return me;
}



xmlDocPtr OCS_request(OCS_t* me, const xmlDocPtr request)
{
	struct curl_slist* http_headers;
	CURLcode curl_res;
	char* xmldoc;
	int   xmldoc_len;
	char* req; 
	size_t req_len;
	char URL[256];
	int ret;
	xmlDocPtr resp_doc;

	xmlParserCtxtPtr parser;
	z_stream z_strm;



	assert (me!= NULL);

	/* compress & dump request document */
	xmlDocDumpFormatMemory (request, (xmlChar**) &xmldoc, &xmldoc_len, 1);
	assert (xmldoc_len != 0);

	req_len = doc_compress ( xmldoc, xmldoc_len, &req);
	assert (req_len > 0);  

	/* setup XML parser and zlib structures */

	parser = xmlCreatePushParserCtxt(NULL, NULL, NULL, 0, NULL);
	assert (parser != NULL);
	parser->_private = &z_strm;

	memset(&z_strm, 0, sizeof (z_stream));
	z_strm.zalloc = Z_NULL;
	z_strm.zfree = Z_NULL;
	z_strm.opaque = Z_NULL;

	ret = inflateInit (&z_strm);
	assert (ret == Z_OK);


	/* setup CURL options */

#if LIBCURL_VERSION_NUM > 0x070c01
        curl_easy_reset (me->curl);     /* available since 7.12.1 */
#else
        curl_easy_cleanup (me->curl);
        me->curl = curl_easy_init ();
        assert (me->curl != NULL);
#endif

	snprintf(URL, sizeof (URL), "http://%s//ocsinventory", me->server);
	curl_easy_setopt (me->curl, CURLOPT_URL, URL);  

	if (debug) curl_easy_setopt (me->curl, CURLOPT_VERBOSE, 1);
	curl_easy_setopt (me->curl, CURLOPT_DEBUGFUNCTION, debug_print_cb );
	curl_easy_setopt (me->curl, CURLOPT_USERAGENT, AGENT_NAME "/" AGENT_VERSION);  
	curl_easy_setopt (me->curl, CURLOPT_FOLLOWLOCATION, 1);
	curl_easy_setopt (me->curl, CURLOPT_POST, 1);


	http_headers = curl_slist_append (NULL, "Content-Type: application/x-compress");
	http_headers = curl_slist_append (http_headers, "Pragma: no-cache");

	curl_easy_setopt (me->curl, CURLOPT_HTTPHEADER, http_headers);
	curl_easy_setopt (me->curl, CURLOPT_POSTFIELDS, req);
	curl_easy_setopt (me->curl, CURLOPT_POSTFIELDSIZE, req_len);	

	curl_easy_setopt (me->curl, CURLOPT_WRITEFUNCTION, write_doc_cb);	
	curl_easy_setopt (me->curl, CURLOPT_WRITEDATA, parser);	  

	curl_res = curl_easy_perform (me->curl);
	inflateEnd(&z_strm);

	if (curl_res == 0)
	{
		xmlParseChunk(parser, NULL, 0, 1);
		resp_doc = parser->myDoc;
	}
	else
	{
#if LIBCURL_VERSION_NUM > 0x070c00
                LOG("** ERROR: %s", curl_easy_strerror(curl_res));
#else
                LOG("** ERROR: %i", curl_res);
#endif
		resp_doc = NULL;
	}

	xmlFreeParserCtxt(parser);
	xmlFree (xmldoc);
	free (req);
	curl_slist_free_all (http_headers);

	return resp_doc;
}

void OCS_free (OCS_t* me)
{
	if (!me) return;

	if (me->curl) 
		curl_easy_cleanup(me->curl);

	if (me->server) 
		free (me->server);

	free (me);

}


